@@ -31,12 +31,12 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: irc.c,v 1.117 2004/12/19 19:25:16 f Exp $
+ * $Id: irc.c,v 1.125 2006/04/30 14:20:56 f Exp $
  */
 
-#define IRCII_VERSION	"20041219"	/* YYYYMMDD */
+#define IRCII_VERSION	"20060501"	/* YYYYMMDD */
 /**************************** PATCHED by Flier ******************************/
-#define SCROLLZ_VERSION "1.9.4"
+#define SCROLLZ_VERSION "1.9.98"
 /****************************************************************************/
 
 #include "irc.h"
@@ -111,7 +111,7 @@ int	irc_port = IRC_PORT,			/* port of ircd */
 	who_mask = 0;				/* keeps track of which /who
 						 * switchs are set */
 
-char	*zero	= "0";
+char	*irczero	= "0";
 char	*one	= "1";
 
 char	oper_command = 0;	/* true just after an oper() command is
@@ -265,7 +265,7 @@ char  defban;
 char  bold=2;
 char  *DefaultServer=(char *) 0;
 char  *ScrollZstr=(char *) 0;
-char  *ScrollZver="ircII " IRCII_VERSION "+ScrollZ " SCROLLZ_VERSION " (19.12.2004)+Cdcc v" CDCC_VERSION;
+char  *ScrollZver="ircII " IRCII_VERSION "+ScrollZ " SCROLLZ_VERSION " (01.05.2006)+Cdcc v" CDCC_VERSION;
 char  *ScrollZver1=(char *) 0;
 #ifdef EXTRA_STUFF
 char  *EString=(char *) 0;
@@ -354,6 +354,11 @@ char  *ChanLogChannels=(char *) 0;
 char  *ChanLogDir=(char *) 0;
 char  *ChanLogPrefix=(char *) 0;
 char  *ChanLogPostfix=(char *) 0;
+char  *ExtTopicDelimiter=(char *) 0;
+#ifdef BLAXTHOS
+char  *OperNick=(char *) 0;
+char  *OperPassword=(char *) 0;
+#endif
 int   DeopPrinted;
 int   KickPrinted;
 int   NickPrinted;
@@ -567,6 +572,11 @@ char VersionInfo[] = {
 #else
     'z',
 #endif
+#ifdef BLAXTHOS
+    'B',
+#else
+    'b',
+#endif
 #ifdef OPER
     ' ',
     'O',
@@ -579,6 +589,14 @@ char VersionInfo[] = {
     'S',
     'S',
     'L',
+#endif
+#ifdef HAVE_ICONV_H
+    ' ',
+    'U',
+    'T',
+    'F',
+    '-',
+    '8',
 #endif
     '\0'};
 /*****************************/
@@ -1735,6 +1753,9 @@ main(argc, argv, envp)
 /**************************** Patched by Flier ******************************/
         memset(&forced_ip_addr, 0, sizeof(struct in_addr));
         memset(&MyHostAddr, 0, sizeof(struct in_addr));
+#ifdef HAVE_SSL
+        gnutls_global_init();
+#endif
 /****************************************************************************/
 #ifdef _Windows
 	reset_pointers();
